# --------------------------------------
# HEADER
# --------------------------------------

__author__ = "Tom Bochynek"
__contact__ = "Thomas.Bochynek@monash.edu"

'''
calls posExtractor every 30 seconds, with arguments video name and extraction time.
stores data in - and retrieves it from - logfile.
'''

# --------------------------------------
# IMPORTS
# --------------------------------------

import cv2
import csv
import numpy as np
import os

# --------------------------------------
# GLOBAL VARIABLES
# --------------------------------------

# select debug mode

DEBUG = 0

# select system, for path variables

SYSTEM = 'mac'
# SYSTEM = 'linux'

# data file directories

if SYSTEM == 'mac':
    VID_LIST_CSV = '/Users/bo/Uni/PhD Data/Code/python encounter counter/videoList5x8_mac.csv'  # video names for each experiment
    LOCATIONS_CSV = '/Users/bo/Uni/PhD Data/Code/python encounter counter/position extraction/coords.log'  # flag locations
    PIXEL_SUMS_CSV = '/Users/bo/Uni/PhD Data/Code/python encounter counter/results/pixel_sums.csv'
    PIXEL_SUMS_DETAIL_CSV = '/Users/bo/Uni/PhD Data/Code/python encounter counter/results/pixel_sums_detail.csv'  # storing files after each summing
    PATH_RESULTS = '/Users/bo/Uni/PhD Data/Code/python encounter counter/results/'
    DIFF_IMG_FOLDERS_PATH = "/Users/bo/Uni/PhD Data/Code/python encounter counter/results/diff_img/"


if SYSTEM == 'linux':
    VID_LIST_CSV = '/home/bo/Projects/encounter_counter/encounter-detection/videoList5x8_linux.csv'  # video names for each experiment
    LOCATIONS_CSV = '/home/bo/Projects/encounter_counter/encounter-detection/position extraction/coords.log'  # flag locations
    PIXEL_SUMS_CSV = '/home/bo/Projects/encounter_counter/encounter-detection/results/pixel_sums.csv'
    PIXEL_SUMS_DETAIL_CSV = '/home/bo/Projects/encounter_counter/encounter-detection/results/pixel_sums_detail.csv'  # storing files after each summing
    PATH_RESULTS = '/home/bo/Projects/encounter_counter/encounter-detection/results/'


# video parameters
VID_LENGTH = 1063  # seconds in a video
EXP_DURATION = 3600  # seconds in an experiment
FPS = 23.967
VID_ERROR_FIX = 94 / FPS  # time until end of empty frames at beginning of video

# image processing parameters
SEQ_DURATION = 30  # length of time step in seconds
GRAY_VAL_DIFF = 30  # threshold for comparing pixel values at t and t-1
CROP_OFFSET = 80  # cropping offset relative to path width
MASK_RADIUS = 20  # radius of mask
MARKER_RADIUS = 4 # radius of marker indicating mask origin


# --------------------------------------
# FUNCTIONS
# --------------------------------------

# data handling


class Entry:
    """ stores recorded data with called parameters"""

    def __init__(self, experiment_nr, abs_time, time, path_coords, flag_coords):
        self.experiment_nr = experiment_nr
        self.abs_time = abs_time
        self.time = time
        self.path_coords = path_coords
        self.flag_coords = flag_coords

    def __str__(self):
        path_list = zip(*self.path_coords)

        path_xs = path_list[0]
        path_ys = path_list[1]

        flag_xs = []
        flag_ys = []

        if len(self.flag_coords) > 0:
            flag_list = zip(*self.flag_coords)
            flag_xs = flag_list[0]
            flag_ys = flag_list[1]

        def to_string(list):
            return ' '.join([str(x) for x in list])

        return "%s,%s,%s,%s,%s,%s,%s" % (self.experiment_nr,
                                         self.abs_time,
                                         self.time,
                                         to_string(path_xs),
                                         to_string(path_ys),
                                         to_string(flag_xs),
                                         to_string(flag_ys)
                                         )


def import_vid_list():
    """
    builds list of video names for all experiments from csv
    :return: list of video names
    """

    vid_list = []
    with open(VID_LIST_CSV, 'rb') as f:
        list = csv.reader(f, delimiter=',')
        for row in list:
            vid_list.append(row)
    return vid_list


def import_log_file():
    """
    imports log file
    :return: log file as list of Entry
    """

    entries = []

    with open(LOCATIONS_CSV, 'r') as inputfile:
        for line in inputfile.readlines():
            entries.append(string_to_entry(line))

    return entries


def string_to_entry(string):
    """
    reads string from log file and turns it into Entry
    :param string: string from log file
    :return: Entry object
    """

    string_list = string.split(',')

    experiment_nr = int(string_list[0])
    abs_time = int(string_list[1])
    time = int(string_list[2])

    def to_array(string):
        """
        converts a string into an array
        :param string: Entry-string
        :return: array
        """

        return [int(x) for x in string.split(' ')]

    path_coords = zip(to_array(string_list[3]), to_array(string_list[4]))

    flag_coords = []

    if len(string_list[5]) > 0 and len(string_list[6]) > 0:
        flag_coords = zip(to_array(string_list[5]), to_array(string_list[6]))

    return Entry(experiment_nr, abs_time, time, path_coords, flag_coords)


# image processing

def all_mark(img, flag_coords, crop_x, crop_y):
        """
        updated flag coordinates to relative crop; applies mask to image.
        :param img: gray-scale 30 sec diff image
        :param flag_coords: location of flags at time of masking
        :param crop_x: x coordinates of cropped section
        :param crop_y: y coordinates of cropped section
        :return: masked image
        """

        for i in range(len(flag_coords)):
            upd_coords = (flag_coords[i][0] - crop_x[0], flag_coords[i][1] - crop_y[0])
            cv2.circle(img, upd_coords, MARKER_RADIUS, 255, -1)

        return img

# def img_aggregator(start_time, vid_path, crop_x, crop_y):
#     """
#     aggregates positions of newly "active" pixels. pixel become active if difference between new and old pixels is
#     larger than GRAY_VAL_DIFF.
#     :param start_time: time of aggregation start in ms
#     :param vid_path: path to video to be opened
#     :return: gray-scale image with pixel value indicating number of changed pixel states over time
#     """
#
#     print "starting at time " + str(int(start_time))
#
#     start_frame = int(start_time * FPS)
#     frame_counter = 0
#
#     # open vid at start time
#     vidcap = cv2.VideoCapture(vid_path)
#     vidcap.set(cv2.CAP_PROP_POS_FRAMES, start_frame)
#
#     # importing first image, cropping
#     success, img_old = vidcap.read()
#     img_old = img_old[crop_y[0]:crop_y[1], crop_x[0]:crop_x[1]]
#     img_old = cv2.cvtColor(img_old, cv2.COLOR_BGR2GRAY)
#
#     # creating empty summary image with size of cropped image
#     dimensions = img_old.shape
#     img_sum = np.zeros(dimensions)
#
#     print
#
#     norm_duration = SEQ_DURATION * FPS
#     # multiplier = 1  # in the following lines, fixing problem with video switch.
#
#     if start_time >= 1033:  # allowing for end-of-video extraction - if sequence would be longer
#         # than video, sequence length is decreased (and later masked pixel sum output scaled to
#         # 30 seconds via multiplier
#
#         norm_duration = 15 * FPS
#         multiplier = 2
#
#     if start_time >= 1050:
#         norm_duration = 7.5 * FPS
#         multiplier = 4
#
#     while frame_counter <= norm_duration:
#
#         print '\r', frame_counter,
#         frame_counter += 1
#
#         success, img_new = vidcap.read()
#         img_new = img_new[crop_y[0]:crop_y[1], crop_x[0]:crop_x[1]]
#         img_new = cv2.cvtColor(img_new, cv2.COLOR_BGR2GRAY)
#
#         # comparing images and noting newly active pixels
#         for r in range(len(img_new)):
#             for c in range(len(img_new[r])):
#                 if img_new[r][c] > img_old[r][c]:
#                     diff = img_new[r][c] - img_old[r][c]
#                     if diff > GRAY_VAL_DIFF:
#                         img_sum[r][c] += 1
#                     else:
#                         pass
#                 else:
#                     pass
#
#         img_old = img_new
#
#     return (img_sum, multiplier)

def masker(img, flag_coords, crop_x, crop_y):
    """
    updated flag coordinates to relative crop; applies mask to image.
    :param img: gray-scale 30 sec diff image
    :param flag_coords: location of flags at time of masking
    :param crop_x: x coordinates of cropped section
    :param crop_y: y coordinates of cropped section
    :return: masked image
    """

    pixel_sums_frame = 0
    pixel_sums_all_flags = []

    for i in range(len(flag_coords)):
        upd_coords = (flag_coords[i][0] - crop_x[0], flag_coords[i][1] - crop_y[0])
        pixel_sums_flag = indiv_mask(img, upd_coords)
        pixel_sums_all_flags.append(pixel_sums_flag) # for details of each frame
        pixel_sums_frame += pixel_sums_flag

    return pixel_sums_frame, pixel_sums_all_flags


def indiv_mask(img, upd_coords):
    """
    returns value for one-circle mask
    """
    mask = np.zeros(img.shape[:2], dtype="uint8")
    cv2.circle(mask, upd_coords, MASK_RADIUS, 255, -1)
    masked_img = cv2.bitwise_and(img, img, mask=mask)
    pixel_val_sum = int(np.sum(masked_img))

    # debugging: showing individual frames
    if DEBUG == 1:
        cv2.imshow('a', masked_img)
        cv2.waitKey()

    return pixel_val_sum

# main executing function

def encounter_extraction():
    """
    main extraction function.
    :return:
    """
    entries = import_log_file()
    vid_list = import_vid_list()
    # img_list = img_lists_loader()

    for experiment_nr in [6]:
        print 'starting extraction on experiment', experiment_nr + 1

        # define ROI
        path_coords = entries[(experiment_nr + 1) * 120].path_coords
        print 'path coords: ', path_coords

        # define cropping area
        crop_x = sorted([path_coords[0][0], path_coords[1][0]])
        crop_y = sorted([(path_coords[0][1] - CROP_OFFSET), (path_coords[1][1] + CROP_OFFSET)])

        # print crop_x
        # print crop_y
        abs_time = 0

        # priming empty list of pixel val sums with experiment nr
        details = [experiment_nr + 1]
        exp_sum_list = [experiment_nr + 1]

        # loop over experiment time
        while abs_time <= EXP_DURATION:
            experiment_path = vid_list[experiment_nr][1]
            vid_number = int(abs_time / VID_LENGTH)
            vid_name = vid_list[experiment_nr][vid_number + 2]
            vid_path = experiment_path + vid_name
            time = abs_time % VID_LENGTH

            flag_coords = entries[(experiment_nr * 121) + abs_time / 30].flag_coords
            print 'now looping, time: ' + str(abs_time)
            print flag_coords

            if len(flag_coords) > 0:  # only process for images that have flags on them
                # print flag_coords
                print
                print 'Exp', experiment_nr + 1, 'abs time:', abs_time

                # build filename string
                diff_img_path = DIFF_IMG_FOLDERS_PATH + str(experiment_nr + 1) + '/' + str(experiment_nr + 1) + '_' \
                                + str(abs_time) + '_diff.jpg'

                print diff_img_path

                # load difference image
                diff_img = cv2.imread(diff_img_path)

                if DEBUG == 1:
                    # print str(experiment_nr +1) + ', ' + str(abs_time)
                    cv2.imshow('a', diff_img)
                    cv2.waitKey()

                # masking and summing individual masks
                (mask_sum, mask_details) = masker(diff_img, flag_coords, crop_x, crop_y)

                # collective mask
                all_mark_img = all_mark(diff_img, flag_coords, crop_x, crop_y)

                # writing detail list
                details.extend(mask_details)

                with open(PIXEL_SUMS_DETAIL_CSV, 'ab') as csv_file:
                    writer = csv.writer(csv_file, delimiter=',')
                    writer.writerow(details)

                # growing exp_list
                exp_sum_list.append(mask_sum)

                # # saving diff_img
                # foldername_diff = PATH_RESULTS + 'diff_img/' + str(experiment_nr + 1) + '/'
                #
                # if not os.path.exists(foldername_diff):
                #     os.makedirs(foldername_diff)

                # filename_diff = str(experiment_nr + 1) + '_' + str(abs_time) + '_' + 'diff.jpg'
                #
                # cv2.imwrite(foldername_diff + filename_diff, diff_img)

                # # saving all_mark_img
                foldername_mark = PATH_RESULTS + 'marked_img/' + str(experiment_nr + 1) + '/'

                if not os.path.exists(foldername_mark):
                    os.makedirs(foldername_mark)

                filename_marked = str(experiment_nr + 1) + '_' + str(abs_time) + '_' \
                                + str(mask_sum) + '_' + 'marked.jpg'
                cv2.imwrite(foldername_mark + filename_marked, all_mark_img)

                # # create copy with marked flag_pos
                # marked_diff_img = diff_img
                #
                # for i in range(len(flag_coords)):
                #     upd_coords = (flag_coords[i][0] - crop_x[0], flag_coords[i][1] - crop_y[0])
                #     cv2.circle(marked_diff_img, upd_coords, 3, 255, -1)
                #
                # # saving marked_diff_img
                # foldername_marked_diff = PATH_RESULTS + 'marked_diff_img/' + str(experiment_nr + 1) + '/'
                #
                # if not os.path.exists(foldername_marked_diff):
                #     os.makedirs(foldername_marked_diff)
                #
                # filename_marked_diff = str(experiment_nr + 1) + '_' + str(abs_time) + '_' + 'marked_diff.jpg'
                #
                # cv2.imwrite(foldername_marked_diff + filename_marked_diff, marked_diff_img)

            abs_time += SEQ_DURATION

        # write total experiment results (exp_sum_list) to file
        with open(PIXEL_SUMS_CSV, 'ab') as csv_file:
            writer = csv.writer(csv_file, delimiter=',')
            writer.writerow(exp_sum_list)

        experiment_nr += 1

    print ''
    print "Extraction complete."


# --------------------------------------
# EXECUTION
# --------------------------------------

encounter_extraction()
